@@ -23,6 +23,9 @@ static int rb_thread_critical; /* dummy */
 #include "st.h"
 #endif
 
+#undef RUBY_UNTYPED_DATA_WARNING
+#define RUBY_UNTYPED_DATA_WARNING 1
+
 #if !defined(RHASH_TBL)
 #define RHASH_TBL(h) (RHASH(h)->tbl)
 #endif
@@ -34,11 +37,25 @@ static int rb_thread_critical; /* dummy */
 #define RARRAY_PTR(s) (RARRAY(s)->ptr)
 #define RARRAY_LEN(s) (RARRAY(s)->len)
 #endif
+#if !defined(RARRAY_CONST_PTR)
+#define RARRAY_CONST_PTR(s) (const VALUE *)RARRAY_PTR(s)
+#endif
+#if !defined(RARRAY_AREF)
+#define RARRAY_AREF(a, i) RARRAY_CONST_PTR(a)[i]
+#endif
 
 #if defined(HAVE_STRNDUP) && !defined(_GNU_SOURCE)
 extern char *strndup(const char* _ptr, size_t _len);
 #endif
 
+#ifndef HAVE_RB_SYM2STR
+# define rb_sym2str(obj) rb_id2str(SYM2ID(obj))
+#endif
+
+#ifndef HAVE_RB_ID2STR
+# define rb_id2str(id) rb_str_new2(rb_id2name(id))
+#endif
+
 static VALUE cMethod;
 
 static VALUE cTclTkLib;
@@ -76,6 +93,40 @@ static unsigned long CALLBACK_ID_NUM = 0;
 
 /*************************************/
 
+#ifndef HAVE_STRNDUP
+static char * strndup _((const char *, size_t));
+static char *
+strndup(ptr, len)
+    const char *ptr;
+    size_t len;
+{
+    char *newptr = malloc(len + 1);
+    if (newptr) {
+	memcpy(newptr, ptr, len);
+	newptr[len] = '\0';
+    }
+    return newptr;
+}
+#endif
+
+#ifndef HAVE_RB_ARY_CAT
+static VALUE rb_ary_cat _((VALUE, const VALUE *, long));
+static VALUE
+rb_ary_cat(ary, argv, len)
+    VALUE ary;
+    const VALUE *argv;
+    long len;
+{
+    long i;
+    for (i = 0; i < len; i++) {
+	rb_ary_push(ary, argv[i]);
+    }
+    return ary;
+}
+#endif
+
+/*************************************/
+
 #if defined(HAVE_RB_OBJ_INSTANCE_EXEC) && !defined(RUBY_VM)
 extern VALUE rb_obj_instance_exec _((int, VALUE*, VALUE));
 #endif
@@ -136,7 +187,7 @@ tk_eval_cmd(argc, argv, self)
     VALUE argv[];
     VALUE self;
 {
-    volatile VALUE cmd, rest;
+    VALUE cmd, rest;
 
     rb_scan_args(argc, argv, "1*", &cmd, &rest);
     return rb_eval_cmd(cmd, rest, 0);
@@ -169,7 +220,7 @@ tk_install_cmd_core(cmd)
     volatile VALUE id_num;
 
     id_num = ULONG2NUM(CALLBACK_ID_NUM++);
-    id_num = rb_funcall(id_num, ID_to_s, 0, 0);
+    id_num = rb_funcallv(id_num, ID_to_s, 0, 0);
     id_num = rb_str_append(rb_str_new2(cmd_id_prefix), id_num);
     rb_hash_aset(CALLBACK_TABLE, id_num, cmd);
     return rb_str_append(rb_str_new2(cmd_id_head), id_num);
@@ -266,7 +317,7 @@ to_strkey(key, value, hash)
     VALUE value;
     VALUE hash;
 {
-    rb_hash_aset(hash, rb_funcall(key, ID_to_s, 0, 0), value);
+    rb_hash_aset(hash, rb_funcallv(key, ID_to_s, 0, 0), value);
     return ST_CHECK;
 }
 
@@ -303,10 +354,10 @@ ary2list(ary, enc_flag, self)
     volatile VALUE dst;
     volatile VALUE sys_enc, dst_enc, str_enc;
 
-    sys_enc = rb_funcall(cTclTkLib, ID_encoding, 0, 0);
+    sys_enc = rb_funcallv(cTclTkLib, ID_encoding, 0, 0);
     if (NIL_P(sys_enc)) {
-      sys_enc = rb_funcall(cTclTkLib, ID_encoding_system, 0, 0);
-      sys_enc = rb_funcall(sys_enc, ID_to_s, 0, 0);
+      sys_enc = rb_funcallv(cTclTkLib, ID_encoding_system, 0, 0);
+      sys_enc = rb_funcallv(sys_enc, ID_to_s, 0, 0);
     }
 
     if (NIL_P(enc_flag)) {
@@ -316,15 +367,15 @@ ary2list(ary, enc_flag, self)
         dst_enc = enc_flag;
         req_chk_flag = 0;
     } else {
-        dst_enc = rb_funcall(enc_flag, ID_to_s, 0, 0);
+        dst_enc = rb_funcallv(enc_flag, ID_to_s, 0, 0);
         req_chk_flag = 0;
     }
 
     /* size = RARRAY_LEN(ary); */
     size = 0;
     for(idx = 0; idx < RARRAY_LEN(ary); idx++) {
-        if (RB_TYPE_P(RARRAY_PTR(ary)[idx], T_HASH)) {
-            size += 2 * RHASH_SIZE(RARRAY_PTR(ary)[idx]);
+        if (RB_TYPE_P(RARRAY_AREF(ary, idx), T_HASH)) {
+            size += 2 * RHASH_SIZE(RARRAY_AREF(ary, idx));
         } else {
             size++;
         }
@@ -332,7 +383,7 @@ ary2list(ary, enc_flag, self)
 
     dst = rb_ary_new2(size);
     for(idx = 0; idx < RARRAY_LEN(ary); idx++) {
-        val = RARRAY_PTR(ary)[idx];
+        val = RARRAY_AREF(ary, idx);
         str_val = Qnil;
         switch(TYPE(val)) {
         case T_ARRAY:
@@ -342,7 +393,7 @@ ary2list(ary, enc_flag, self)
             if (req_chk_flag) {
                 str_enc = rb_ivar_get(str_val, ID_at_enc);
                 if (!NIL_P(str_enc)) {
-                    str_enc = rb_funcall(str_enc, ID_to_s, 0, 0);
+                    str_enc = rb_funcallv(str_enc, ID_to_s, 0, 0);
                 } else {
                     str_enc = sys_enc;
                 }
@@ -363,7 +414,7 @@ ary2list(ary, enc_flag, self)
             }
             size2 = RARRAY_LEN(val);
             for(idx2 = 0; idx2 < size2; idx2++) {
-                val2 = RARRAY_PTR(val)[idx2];
+                val2 = RARRAY_AREF(val, idx2);
                 switch(TYPE(val2)) {
                 case T_ARRAY:
                     str_val = ary2list(val2, enc_flag, self);
@@ -389,7 +440,7 @@ ary2list(ary, enc_flag, self)
                 if (req_chk_flag) {
                     str_enc = rb_ivar_get(str_val, ID_at_enc);
                     if (!NIL_P(str_enc)) {
-                        str_enc = rb_funcall(str_enc, ID_to_s, 0, 0);
+                        str_enc = rb_funcallv(str_enc, ID_to_s, 0, 0);
                     } else {
                         str_enc = sys_enc;
                     }
@@ -409,7 +460,7 @@ ary2list(ary, enc_flag, self)
                 if (req_chk_flag) {
                     str_enc = rb_ivar_get(str_val, ID_at_enc);
                     if (!NIL_P(str_enc)) {
-                        str_enc = rb_funcall(str_enc, ID_to_s, 0, 0);
+                        str_enc = rb_funcallv(str_enc, ID_to_s, 0, 0);
                     } else {
                         str_enc = sys_enc;
                     }
@@ -424,13 +475,13 @@ ary2list(ary, enc_flag, self)
 
     if (RTEST(dst_enc) && !NIL_P(sys_enc)) {
         for(idx = 0; idx < RARRAY_LEN(dst); idx++) {
-            str_val = RARRAY_PTR(dst)[idx];
+            str_val = RARRAY_AREF(dst, idx);
             if (rb_obj_respond_to(self, ID_toUTF8, Qtrue)) {
                 str_val = rb_funcall(self, ID_toUTF8, 1, str_val);
             } else {
                 str_val = rb_funcall(cTclTkLib, ID_toUTF8, 1, str_val);
             }
-            RARRAY_PTR(dst)[idx] = str_val;
+            RARRAY_ASET(dst, idx, str_val);
         }
         val = rb_apply(cTclTkLib, ID_merge_tklist, dst);
         if (RB_TYPE_P(dst_enc, T_STRING)) {
@@ -457,10 +508,10 @@ ary2list2(ary, enc_flag, self)
     volatile VALUE dst;
     volatile VALUE sys_enc, dst_enc, str_enc;
 
-    sys_enc = rb_funcall(cTclTkLib, ID_encoding, 0, 0);
+    sys_enc = rb_funcallv(cTclTkLib, ID_encoding, 0, 0);
     if (NIL_P(sys_enc)) {
-      sys_enc = rb_funcall(cTclTkLib, ID_encoding_system, 0, 0);
-      sys_enc = rb_funcall(sys_enc, ID_to_s, 0, 0);
+      sys_enc = rb_funcallv(cTclTkLib, ID_encoding_system, 0, 0);
+      sys_enc = rb_funcallv(sys_enc, ID_to_s, 0, 0);
     }
 
     if (NIL_P(enc_flag)) {
@@ -470,14 +521,14 @@ ary2list2(ary, enc_flag, self)
         dst_enc = enc_flag;
         req_chk_flag = 0;
     } else {
-        dst_enc = rb_funcall(enc_flag, ID_to_s, 0, 0);
+        dst_enc = rb_funcallv(enc_flag, ID_to_s, 0, 0);
         req_chk_flag = 0;
     }
 
     size = RARRAY_LEN(ary);
     dst = rb_ary_new2(size);
     for(idx = 0; idx < RARRAY_LEN(ary); idx++) {
-        val = RARRAY_PTR(ary)[idx];
+        val = RARRAY_AREF(ary, idx);
         str_val = Qnil;
         switch(TYPE(val)) {
         case T_ARRAY:
@@ -504,7 +555,7 @@ ary2list2(ary, enc_flag, self)
             if (req_chk_flag) {
                 str_enc = rb_ivar_get(str_val, ID_at_enc);
                 if (!NIL_P(str_enc)) {
-                    str_enc = rb_funcall(str_enc, ID_to_s, 0, 0);
+                    str_enc = rb_funcallv(str_enc, ID_to_s, 0, 0);
                 } else {
                     str_enc = sys_enc;
                 }
@@ -518,13 +569,13 @@ ary2list2(ary, enc_flag, self)
 
     if (RTEST(dst_enc) && !NIL_P(sys_enc)) {
         for(idx = 0; idx < RARRAY_LEN(dst); idx++) {
-            str_val = RARRAY_PTR(dst)[idx];
+            str_val = RARRAY_AREF(dst, idx);
             if (rb_obj_respond_to(self, ID_toUTF8, Qtrue)) {
                 str_val = rb_funcall(self, ID_toUTF8, 1, str_val);
             } else {
                 str_val = rb_funcall(cTclTkLib, ID_toUTF8, 1, str_val);
             }
-            RARRAY_PTR(dst)[idx] = str_val;
+            RARRAY_ASET(dst, idx, str_val);
         }
         val = rb_apply(cTclTkLib, ID_merge_tklist, dst);
         if (RB_TYPE_P(dst_enc, T_STRING)) {
@@ -543,7 +594,7 @@ static VALUE
 key2keyname(key)
     VALUE key;
 {
-    return rb_str_append(rb_str_new2("-"), rb_funcall(key, ID_to_s, 0, 0));
+    return rb_str_append(rb_str_new2("-"), rb_funcallv(key, ID_to_s, 0, 0));
 }
 
 static VALUE
@@ -552,7 +603,7 @@ assoc2kv(assoc, ary, self)
     VALUE ary;
     VALUE self;
 {
-    long i, j, len;
+    long i, len;
     volatile VALUE pair;
     volatile VALUE val;
     volatile VALUE dst = rb_ary_new2(2 * RARRAY_LEN(assoc));
@@ -560,28 +611,26 @@ assoc2kv(assoc, ary, self)
     len = RARRAY_LEN(assoc);
 
     for(i = 0; i < len; i++) {
-        pair = RARRAY_PTR(assoc)[i];
+        pair = RARRAY_AREF(assoc, i);
         if (!RB_TYPE_P(pair, T_ARRAY)) {
             rb_ary_push(dst, key2keyname(pair));
             continue;
         }
         switch(RARRAY_LEN(assoc)) {
         case 2:
-            rb_ary_push(dst, RARRAY_PTR(pair)[2]);
+            rb_ary_push(dst, RARRAY_AREF(pair, 2));
 
         case 1:
-            rb_ary_push(dst, key2keyname(RARRAY_PTR(pair)[0]));
+            rb_ary_push(dst, key2keyname(RARRAY_AREF(pair, 0)));
 
         case 0:
             continue;
 
         default:
-            rb_ary_push(dst, key2keyname(RARRAY_PTR(pair)[0]));
+            rb_ary_push(dst, key2keyname(RARRAY_AREF(pair, 0)));
 
             val = rb_ary_new2(RARRAY_LEN(pair) - 1);
-            for(j = 1; j < RARRAY_LEN(pair); j++) {
-                rb_ary_push(val, RARRAY_PTR(pair)[j]);
-            }
+            rb_ary_cat(val, RARRAY_CONST_PTR(pair) + 1, RARRAY_LEN(pair) - 1);
 
             rb_ary_push(dst, val);
         }
@@ -600,7 +649,7 @@ assoc2kv_enc(assoc, ary, self)
     VALUE ary;
     VALUE self;
 {
-    long i, j, len;
+    long i, len;
     volatile VALUE pair;
     volatile VALUE val;
     volatile VALUE dst = rb_ary_new2(2 * RARRAY_LEN(assoc));
@@ -608,28 +657,26 @@ assoc2kv_enc(assoc, ary, self)
     len = RARRAY_LEN(assoc);
 
     for(i = 0; i < len; i++) {
-        pair = RARRAY_PTR(assoc)[i];
+        pair = RARRAY_AREF(assoc, i);
         if (!RB_TYPE_P(pair, T_ARRAY)) {
             rb_ary_push(dst, key2keyname(pair));
             continue;
         }
         switch(RARRAY_LEN(assoc)) {
         case 2:
-            rb_ary_push(dst, get_eval_string_core(RARRAY_PTR(pair)[2], Qtrue, self));
+            rb_ary_push(dst, get_eval_string_core(RARRAY_AREF(pair, 2), Qtrue, self));
 
         case 1:
-            rb_ary_push(dst, key2keyname(RARRAY_PTR(pair)[0]));
+            rb_ary_push(dst, key2keyname(RARRAY_AREF(pair, 0)));
 
         case 0:
             continue;
 
         default:
-            rb_ary_push(dst, key2keyname(RARRAY_PTR(pair)[0]));
+            rb_ary_push(dst, key2keyname(RARRAY_AREF(pair, 0)));
 
             val = rb_ary_new2(RARRAY_LEN(pair) - 1);
-            for(j = 1; j < RARRAY_LEN(pair); j++) {
-                rb_ary_push(val, RARRAY_PTR(pair)[j]);
-            }
+            rb_ary_cat(val, RARRAY_CONST_PTR(pair) + 1, RARRAY_LEN(pair) - 1);
 
             rb_ary_push(dst, get_eval_string_core(val, Qtrue, self));
         }
@@ -650,7 +697,7 @@ push_kv(key, val, args)
 {
     volatile VALUE ary;
 
-    ary = RARRAY_PTR(args)[0];
+    ary = RARRAY_AREF(args, 0);
 
 #if 0
     rb_ary_push(ary, key2keyname(key));
@@ -660,7 +707,7 @@ push_kv(key, val, args)
 
     if (val == TK_None) return ST_CHECK;
 
-    rb_ary_push(ary, get_eval_string_core(val, Qnil, RARRAY_PTR(args)[1]));
+    rb_ary_push(ary, get_eval_string_core(val, Qnil, RARRAY_AREF(args, 1)));
 
     return ST_CHECK;
 }
@@ -691,20 +738,20 @@ push_kv_enc(key, val, args)
 {
     volatile VALUE ary;
 
-    ary = RARRAY_PTR(args)[0];
+    ary = RARRAY_AREF(args, 0);
 
 #if 0
     rb_ary_push(ary, key2keyname(key));
     if (val != TK_None) {
         rb_ary_push(ary, get_eval_string_core(val, Qtrue,
-                                              RARRAY_PTR(args)[1]));
+                                              RARRAY_AREF(args, 1)));
     }
 #endif
     rb_ary_push(ary, key2keyname(key));
 
     if (val == TK_None) return ST_CHECK;
 
-    rb_ary_push(ary, get_eval_string_core(val, Qtrue, RARRAY_PTR(args)[1]));
+    rb_ary_push(ary, get_eval_string_core(val, Qtrue, RARRAY_AREF(args, 1)));
 
     return ST_CHECK;
 }
@@ -757,6 +804,7 @@ tk_hash_kv(argc, argv, self)
     switch(argc) {
     case 3:
         ary = argv[2];
+	if (!NIL_P(ary)) Check_Type(ary, T_ARRAY);
     case 2:
         enc_flag = argv[1];
     case 1:
@@ -814,7 +862,7 @@ get_eval_string_core(obj, enc_flag, self)
     case T_FLOAT:
     case T_FIXNUM:
     case T_BIGNUM:
-        return rb_funcall(obj, ID_to_s, 0, 0);
+        return rb_funcallv(obj, ID_to_s, 0, 0);
 
     case T_STRING:
         if (RTEST(enc_flag)) {
@@ -836,11 +884,7 @@ get_eval_string_core(obj, enc_flag, self)
                 return fromDefaultEnc_toUTF8(rb_sym2str(obj), self);
             }
         } else {
-#ifdef HAVE_RB_SYM_TO_S
-            return rb_sym_to_s(obj);
-#else
             return rb_sym2str(obj);
-#endif
         }
 
     case T_HASH:
@@ -863,12 +907,12 @@ get_eval_string_core(obj, enc_flag, self)
         return rb_str_new2("");
 
     case T_REGEXP:
-        return rb_funcall(obj, ID_source, 0, 0);
+        return rb_funcallv(obj, ID_source, 0, 0);
 
     default:
         if (rb_obj_is_kind_of(obj, cTkObject)) {
-            /* return rb_str_new3(rb_funcall(obj, ID_path, 0, 0)); */
-            return get_eval_string_core(rb_funcall(obj, ID_path, 0, 0),
+            /* return rb_str_new3(rb_funcallv(obj, ID_path, 0, 0)); */
+            return get_eval_string_core(rb_funcallv(obj, ID_path, 0, 0),
                                         enc_flag, self);
         }
 
@@ -885,15 +929,15 @@ get_eval_string_core(obj, enc_flag, self)
         if (obj == TK_None)  return Qnil;
 
         if (rb_obj_respond_to(obj, ID_to_eval, Qtrue)) {
-            /* return rb_funcall(obj, ID_to_eval, 0, 0); */
-            return get_eval_string_core(rb_funcall(obj, ID_to_eval, 0, 0),
+            /* return rb_funcallv(obj, ID_to_eval, 0, 0); */
+            return get_eval_string_core(rb_funcallv(obj, ID_to_eval, 0, 0),
                                         enc_flag, self);
         } else if (rb_obj_respond_to(obj, ID_path, Qtrue)) {
-            /* return rb_funcall(obj, ID_path, 0, 0); */
-            return get_eval_string_core(rb_funcall(obj, ID_path, 0, 0),
+            /* return rb_funcallv(obj, ID_path, 0, 0); */
+            return get_eval_string_core(rb_funcallv(obj, ID_path, 0, 0),
                                         enc_flag, self);
         } else if (rb_obj_respond_to(obj, ID_to_s, Qtrue)) {
-            return rb_funcall(obj, ID_to_s, 0, 0);
+            return rb_funcallv(obj, ID_to_s, 0, 0);
         }
     }
 
@@ -908,7 +952,7 @@ tk_get_eval_string(argc, argv, self)
     VALUE *argv;
     VALUE self;
 {
-    volatile VALUE obj, enc_flag;
+    VALUE obj, enc_flag;
 
     if (rb_scan_args(argc, argv, "11", &obj, &enc_flag) == 1) {
         enc_flag = Qnil;
@@ -1168,6 +1212,23 @@ subst_free(ptr)
     }
 }
 
+static size_t
+subst_memsize(ptr)
+    const struct cbsubst_info *ptr;
+{
+    return sizeof(*ptr);
+}
+
+static const rb_data_type_t cbsubst_info_type = {
+    "TkUtil/CallbackSubst/Info",
+    {
+	subst_mark,
+	subst_free,
+	subst_memsize,
+    },
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY,
+};
+
 static VALUE
 allocate_cbsubst_info(struct cbsubst_info **inf_ptr)
 {
@@ -1175,7 +1236,8 @@ allocate_cbsubst_info(struct cbsubst_info **inf_ptr)
   volatile VALUE proc, aliases;
   int idx;
 
-  inf = ALLOC(struct cbsubst_info);
+  VALUE info = TypedData_Make_Struct(cSUBST_INFO, struct cbsubst_info,
+				     &cbsubst_info_type, inf);
 
   inf->full_subst_length = 0;
 
@@ -1194,16 +1256,23 @@ allocate_cbsubst_info(struct cbsubst_info **inf_ptr)
 
   if (inf_ptr != (struct cbsubst_info **)NULL) *inf_ptr = inf;
 
-  return Data_Wrap_Struct(cSUBST_INFO, subst_mark, subst_free, inf);
+  return info;
 }
 
 static void
-cbsubst_init()
+cbsubst_init(void)
 {
   rb_const_set(cCB_SUBST, ID_SUBST_INFO,
 	       allocate_cbsubst_info((struct cbsubst_info **)NULL));
 }
 
+static struct cbsubst_info *
+cbsubst_get_ptr(klass)
+    VALUE klass;
+{
+    return rb_check_typeddata(rb_const_get(klass, ID_SUBST_INFO), &cbsubst_info_type);
+}
+
 static VALUE
 cbsubst_initialize(argc, argv, self)
     int   argc;
@@ -1213,14 +1282,15 @@ cbsubst_initialize(argc, argv, self)
     struct cbsubst_info *inf;
     int idx, iv_idx;
 
-    Data_Get_Struct(rb_const_get(rb_obj_class(self), ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(rb_obj_class(self));
 
-   idx = 0;
-    for(iv_idx = 0; iv_idx < CBSUBST_TBL_MAX; iv_idx++) {
-      if ( inf->ivar[iv_idx] == (ID) 0 ) continue;
-      rb_ivar_set(self, inf->ivar[iv_idx], argv[idx++]);
-      if (idx >= argc) break;
+    if (argc > 0) {
+	idx = 0;
+	for (iv_idx = 0; iv_idx < CBSUBST_TBL_MAX; iv_idx++) {
+	    if (inf->ivar[iv_idx] == (ID)0) continue;
+	    rb_ivar_set(self, inf->ivar[iv_idx], argv[idx++]);
+	    if (idx >= argc) break;
+	}
     }
 
     return self;
@@ -1285,14 +1355,46 @@ cbsubst_def_attr_aliases(self, tbl)
         rb_raise(rb_eArgError, "expected a Hash");
     }
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(self);
 
     rb_hash_foreach(tbl, each_attr_def, self);
 
     return rb_funcall(inf->aliases, rb_intern("update"), 1, tbl);
 }
 
+static VALUE
+cbsubst_append_inf_key(str, inf, idx)
+    VALUE str;
+    const struct cbsubst_info *inf;
+    int idx;
+{
+    const long len = inf->keylen[idx];
+    const long olen = RSTRING_LEN(str);
+    char *buf, *ptr;
+
+    rb_str_modify_expand(str, (len ? len : 1) + 2);
+    buf = RSTRING_PTR(str);
+    ptr = buf + olen;
+
+    *(ptr++) = '%';
+
+    if (len != 0) {
+	/* longname */
+	strncpy(ptr, inf->key[idx], len);
+	ptr += len;
+    }
+    else {
+	/* single char */
+	*(ptr++) = (unsigned char)idx;
+    }
+
+    *(ptr++) = ' ';
+
+    rb_str_set_len(str, ptr - buf);
+
+    return str;
+}
+
 static VALUE
 cbsubst_sym_to_subst(self, sym)
     VALUE self;
@@ -1300,16 +1402,13 @@ cbsubst_sym_to_subst(self, sym)
 {
     struct cbsubst_info *inf;
     VALUE str;
-    char *buf, *ptr;
     int idx;
-    long len;
     ID id;
     volatile VALUE ret;
 
     if (!RB_TYPE_P(sym, T_SYMBOL)) return sym;
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(self);
 
     if (!NIL_P(ret = rb_hash_aref(inf->aliases, sym))) {
 	str = rb_sym2str(ret);
@@ -1324,27 +1423,7 @@ cbsubst_sym_to_subst(self, sym)
     }
     if (idx >= CBSUBST_TBL_MAX)  return sym;
 
-    ptr = buf = ALLOC_N(char, inf->full_subst_length + 1);
-
-    *(ptr++) = '%';
-
-    if (len = inf->keylen[idx]) {
-      /* longname */
-      strncpy(ptr, inf->key[idx], len);
-      ptr += len;
-    } else {
-      /* single char */
-      *(ptr++) = (unsigned char)idx;
-    }
-
-    *(ptr++) = ' ';
-    *(ptr++) = '\0';
-
-    ret = rb_str_new2(buf);
-
-    xfree(buf);
-
-    return ret;
+    return cbsubst_append_inf_key(rb_str_new(0, 0), inf, idx);
 }
 
 static VALUE
@@ -1355,17 +1434,13 @@ cbsubst_get_subst_arg(argc, argv, self)
 {
     struct cbsubst_info *inf;
     VALUE str;
-    char *buf, *ptr;
     int i, idx;
-    long len;
     ID id;
-    volatile VALUE arg_sym, ret;
+    VALUE arg_sym, ret, result;
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
-
-    ptr = buf = ALLOC_N(char, inf->full_subst_length + 1);
+    inf = cbsubst_get_ptr(self);
 
+    result = rb_str_new(0, 0);
     for(i = 0; i < argc; i++) {
         switch(TYPE(argv[i])) {
         case T_STRING:
@@ -1397,27 +1472,10 @@ cbsubst_get_subst_arg(argc, argv, self)
             rb_raise(rb_eArgError, "cannot find attribute :%"PRIsVALUE, str);
         }
 
-	*(ptr++) = '%';
-
-	if (len = inf->keylen[idx]) {
-	  /* longname */
-	  strncpy(ptr, inf->key[idx], len);
-	  ptr += len;
-	} else {
-	  /* single char */
-	  *(ptr++) = (unsigned char)idx;
-	}
-
-	*(ptr++) = ' ';
+	result = cbsubst_append_inf_key(result, inf, idx);
     }
 
-    *ptr = '\0';
-
-    ret = rb_str_new2(buf);
-
-    xfree(buf);
-
-    return ret;
+    return result;
 }
 
 static VALUE
@@ -1426,24 +1484,24 @@ cbsubst_get_subst_key(self, str)
     VALUE str;
 {
     struct cbsubst_info *inf;
-    volatile VALUE list;
-    volatile VALUE ret;
-    VALUE keyval;
+    VALUE list;
+    VALUE ret;
     long i, len, keylen;
     int idx;
-    char *buf, *ptr, *key;
+    char *buf, *ptr;
 
     list = rb_funcall(cTclTkLib, ID_split_tklist, 1, str);
+    Check_Type(list, T_ARRAY);
     len = RARRAY_LEN(list);
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(self);
 
-    ptr = buf = ALLOC_N(char, inf->full_subst_length + len + 1);
+    ret = rb_str_new(0, len);
+    ptr = buf = RSTRING_PTR(ret);
 
     for(i = 0; i < len; i++) {
-      keyval = RARRAY_PTR(list)[i];
-      key = RSTRING_PTR(keyval);
+      VALUE keyval = RARRAY_AREF(list, i);
+      const char *key = (Check_Type(keyval, T_STRING), StringValueCStr(keyval));
       if (*key == '%') {
 	if (*(key + 2) == '\0') {
 	  /* single char */
@@ -1467,10 +1525,8 @@ cbsubst_get_subst_key(self, str)
 	*(ptr++) = ' ';
       }
     }
-    *ptr = '\0';
 
-    ret = rb_str_new2(buf);
-    xfree(buf);
+    rb_str_set_len(ret, ptr - buf);
     return ret;
 }
 
@@ -1479,46 +1535,26 @@ cbsubst_get_all_subst_keys(self)
     VALUE self;
 {
     struct cbsubst_info *inf;
-    char *buf, *ptr;
     char *keys_buf, *keys_ptr;
     int idx;
-    long len;
-    volatile VALUE ret;
+    VALUE str, keys_str;
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(self);
 
-    ptr = buf = ALLOC_N(char, inf->full_subst_length + 1);
-    keys_ptr = keys_buf = ALLOC_N(char, CBSUBST_TBL_MAX + 1);
+    str = rb_str_new(0, 0);
+    keys_str = rb_str_new(0, CBSUBST_TBL_MAX);
+    keys_ptr = keys_buf = RSTRING_PTR(keys_str);
 
     for(idx = 0; idx < CBSUBST_TBL_MAX; idx++) {
       if (inf->ivar[idx] == (ID) 0) continue;
 
       *(keys_ptr++) = (unsigned char)idx;
 
-      *(ptr++) = '%';
-
-      if (len = inf->keylen[idx]) {
-	/* longname */
-	strncpy(ptr, inf->key[idx], len);
-	ptr += len;
-      } else {
-	/* single char */
-	*(ptr++) = (unsigned char)idx;
-      }
-
-      *(ptr++) = ' ';
+      str = cbsubst_append_inf_key(str, inf, idx);
     }
+    rb_str_set_len(keys_str, keys_ptr - keys_buf);
 
-    *ptr = '\0';
-    *keys_ptr = '\0';
-
-    ret = rb_ary_new3(2, rb_str_new2(keys_buf), rb_str_new2(buf));
-
-    xfree(buf);
-    xfree(keys_buf);
-
-    return ret;
+    return rb_ary_new3(2, keys_str, str);
 }
 
 static VALUE
@@ -1527,14 +1563,15 @@ cbsubst_table_setup(argc, argv, self)
      VALUE *argv;
      VALUE self;
 {
-  volatile VALUE cbsubst_obj;
-  volatile VALUE key_inf;
-  volatile VALUE longkey_inf;
-  volatile VALUE proc_inf;
-  VALUE inf;
+  VALUE cbsubst_obj;
+  VALUE key_inf;
+  VALUE longkey_inf;
+  VALUE proc_inf;
+  VALUE inf, subst, name, type, ivar, proc;
+  const VALUE *infp;
   ID id;
   struct cbsubst_info *subst_inf;
-  long idx, len;
+  long idx;
   unsigned char chr;
 
   /* accept (key_inf, proc_inf) or (key_inf, longkey_inf, procinf) */
@@ -1542,6 +1579,9 @@ cbsubst_table_setup(argc, argv, self)
     proc_inf = longkey_inf;
     longkey_inf = rb_ary_new();
   }
+  Check_Type(key_inf, T_ARRAY);
+  Check_Type(longkey_inf, T_ARRAY);
+  Check_Type(proc_inf, T_ARRAY);
 
   /* check the number of longkeys */
   if (RARRAY_LEN(longkey_inf) > 125 /* from 0x80 to 0xFD */) {
@@ -1557,29 +1597,26 @@ cbsubst_table_setup(argc, argv, self)
    *         type  ==> char code or string
    *         ivar  ==> symbol
    */
-  len = RARRAY_LEN(key_inf);
-  for(idx = 0; idx < len; idx++) {
-    inf = RARRAY_PTR(key_inf)[idx];
+  for(idx = 0; idx < RARRAY_LEN(key_inf); idx++) {
+    inf = RARRAY_AREF(key_inf, idx);
     if (!RB_TYPE_P(inf, T_ARRAY)) continue;
+    if (RARRAY_LEN(inf) < 3) continue;
+    infp = RARRAY_CONST_PTR(inf);
+    subst = infp[0];
+    type = infp[1];
+    ivar = infp[2];
 
-    if (RB_TYPE_P(RARRAY_PTR(inf)[0], T_STRING)) {
-      chr = *(RSTRING_PTR(RARRAY_PTR(inf)[0]));
-    } else {
-      chr = NUM2CHR(RARRAY_PTR(inf)[0]);
-    }
-    if (RB_TYPE_P(RARRAY_PTR(inf)[1], T_STRING)) {
-      subst_inf->type[chr] = *(RSTRING_PTR(RARRAY_PTR(inf)[1]));
-    } else {
-      subst_inf->type[chr] = NUM2CHR(RARRAY_PTR(inf)[1]);
-    }
+    chr = NUM2CHR(subst);
+    subst_inf->type[chr] = NUM2CHR(type);
 
     subst_inf->full_subst_length += 3;
 
-    id = SYM2ID(RARRAY_PTR(inf)[2]);
+    id = SYM2ID(ivar);
     subst_inf->ivar[chr] = rb_intern_str(rb_sprintf("@%"PRIsVALUE, rb_id2str(id)));
 
     rb_attr(self, id, 1, 0, Qtrue);
   }
+  RB_GC_GUARD(key_inf);
 
 
   /*
@@ -1588,53 +1625,47 @@ cbsubst_table_setup(argc, argv, self)
    *         type ==> char code or string
    *         ivar ==> symbol
    */
-  len = RARRAY_LEN(longkey_inf);
-  for(idx = 0; idx < len; idx++) {
-    inf = RARRAY_PTR(longkey_inf)[idx];
+  for(idx = 0; idx < RARRAY_LEN(longkey_inf); idx++) {
+    inf = RARRAY_AREF(longkey_inf, idx);
     if (!RB_TYPE_P(inf, T_ARRAY)) continue;
+    if (RARRAY_LEN(inf) < 3) continue;
+    infp = RARRAY_CONST_PTR(inf);
+    name = infp[0];
+    type = infp[1];
+    ivar = infp[2];
 
+    Check_Type(name, T_STRING);
     chr = (unsigned char)(0x80 + idx);
-    subst_inf->keylen[chr] = RSTRING_LEN(RARRAY_PTR(inf)[0]);
-#if HAVE_STRNDUP
-    subst_inf->key[chr] = strndup(RSTRING_PTR(RARRAY_PTR(inf)[0]),
-				  RSTRING_LEN(RARRAY_PTR(inf)[0]));
-#else
-    subst_inf->key[chr] = malloc(RSTRING_LEN(RARRAY_PTR(inf)[0]) + 1);
-    if (subst_inf->key[chr]) {
-      strncpy(subst_inf->key[chr], RSTRING_PTR(RARRAY_PTR(inf)[0]),
-	      RSTRING_LEN(RARRAY_PTR(inf)[0]) + 1);
-      subst_inf->key[chr][RSTRING_LEN(RARRAY_PTR(inf)[0])] = '\0';
-    }
-#endif
-    if (RB_TYPE_P(RARRAY_PTR(inf)[1], T_STRING)) {
-      subst_inf->type[chr] = *(RSTRING_PTR(RARRAY_PTR(inf)[1]));
-    } else {
-      subst_inf->type[chr] = NUM2CHR(RARRAY_PTR(inf)[1]);
-    }
+    subst_inf->keylen[chr] = RSTRING_LEN(name);
+    subst_inf->key[chr] = strndup(RSTRING_PTR(name),
+				  RSTRING_LEN(name));
+    subst_inf->type[chr] = NUM2CHR(type);
 
     subst_inf->full_subst_length += (subst_inf->keylen[chr] + 2);
 
-    id = SYM2ID(RARRAY_PTR(inf)[2]);
+    id = SYM2ID(ivar);
     subst_inf->ivar[chr] = rb_intern_str(rb_sprintf("@%"PRIsVALUE, rb_id2str(id)));
 
     rb_attr(self, id, 1, 0, Qtrue);
   }
+  RB_GC_GUARD(longkey_inf);
 
   /*
    * procs : array of [type, proc]
    *         type  ==> char code or string
    *         proc  ==> proc/method/obj (must respond to 'call')
    */
-  len = RARRAY_LEN(proc_inf);
-  for(idx = 0; idx < len; idx++) {
-    inf = RARRAY_PTR(proc_inf)[idx];
+  for(idx = 0; idx < RARRAY_LEN(proc_inf); idx++) {
+    inf = RARRAY_AREF(proc_inf, idx);
     if (!RB_TYPE_P(inf, T_ARRAY)) continue;
-    rb_hash_aset(subst_inf->proc,
-		 (RB_TYPE_P(RARRAY_PTR(inf)[0], T_STRING)?
-		  INT2FIX(*(RSTRING_PTR(RARRAY_PTR(inf)[0]))) :
-		  RARRAY_PTR(inf)[0]),
-		 RARRAY_PTR(inf)[1]);
+    if (RARRAY_LEN(inf) < 2) continue;
+    type = rb_ary_entry(inf, 0);
+    proc = rb_ary_entry(inf, 1);
+    if (RB_TYPE_P(type, T_STRING))
+      type = INT2FIX(*(RSTRING_PTR(type)));
+    rb_hash_aset(subst_inf->proc, type, proc);
   }
+  RB_GC_GUARD(proc_inf);
 
   rb_const_set(self, ID_SUBST_INFO, cbsubst_obj);
 
@@ -1656,9 +1687,9 @@ cbsubst_scan_args(self, arg_key, val_ary)
 {
     struct cbsubst_info *inf;
     long idx;
-    unsigned char *keyptr = (unsigned char*)RSTRING_PTR(arg_key);
+    unsigned char *keyptr = (unsigned char*)StringValueCStr(arg_key);
     long keylen = RSTRING_LEN(arg_key);
-    long vallen = RARRAY_LEN(val_ary);
+    long vallen = (Check_Type(val_ary, T_ARRAY), RARRAY_LEN(val_ary));
     unsigned char type_chr;
     volatile VALUE dst = rb_ary_new2(vallen);
     volatile VALUE proc;
@@ -1670,8 +1701,7 @@ cbsubst_scan_args(self, arg_key, val_ary)
 
     old_gc = rb_gc_disable();
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(self);
 
     for(idx = 0; idx < vallen; idx++) {
       if (idx >= keylen) {
@@ -1679,7 +1709,7 @@ cbsubst_scan_args(self, arg_key, val_ary)
       } else if (*(keyptr + idx) == ' ') {
 	proc = Qnil;
       } else {
-	if (type_chr = inf->type[*(keyptr + idx)]) {
+	if ((type_chr = inf->type[*(keyptr + idx)]) != 0) {
 	  proc = rb_hash_aref(inf->proc, INT2FIX((int)type_chr));
 	} else {
 	  proc = Qnil;
@@ -1687,10 +1717,10 @@ cbsubst_scan_args(self, arg_key, val_ary)
       }
 
       if (NIL_P(proc)) {
-	rb_ary_push(dst, RARRAY_PTR(val_ary)[idx]);
+	rb_ary_push(dst, RARRAY_AREF(val_ary, idx));
       } else {
 	rb_ary_push(dst, rb_funcall(proc, ID_call, 1,
-				    RARRAY_PTR(val_ary)[idx]));
+				    RARRAY_AREF(val_ary, idx)));
       }
     }
 
@@ -1738,7 +1768,7 @@ tkobj_path(self)
 const char tkutil_release_date[] = TKUTIL_RELEASE_DATE;
 
 void
-Init_tkutil()
+Init_tkutil(void)
 {
     VALUE cTK = rb_define_class("TkKernel", rb_cObject);
     VALUE mTK = rb_define_module("TkUtil");
